// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <sys/io.h>
#include <time.h>

#include "msg.h"
#include "furnace.h"
#include "vars.h"

#define MHz 200
#define NAVG 8

void upause(unsigned);

extern __inline__ unsigned long long int
rdtsc()
{
  unsigned long long int x;
  __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
  return x;
}

static unsigned long long time_base=0;

void
zero_time()
{
  time_base = rdtsc();
}

unsigned int
elapsed_time()
{
  return (int)((rdtsc() - time_base) / MHz);
}


char *ts()
{
  static char buf[40];
  time_t now;
  struct tm *t;
  time(&now);
  t = localtime(&now);
  sprintf(buf, "%02d:%02d:%02d", t->tm_hour, t->tm_min, t->tm_sec);
  return buf;
}

typedef struct {
  unsigned char temp_ser[8];
  unsigned char button_ser[4][8];
  double temp_addend, humidity_addend;
  double temp;
  double humidity;
  unsigned char buttons;
  unsigned char old_buttons;
  int avgidx;
  double temps[NAVG];
  double humids[NAVG];
} ZoneStruct;

ZoneStruct Zone[] = {
  {
    /* This one is marked "1" */
    { 0x26, 0x24, 0x59, 0x0d, 0x00, 0x00, 0x00, 0xf0 },
    {
      { 0x01, 0xe2, 0x8b, 0xb3, 0x07, 0x00, 0x00, 0xa2 },
      { 0x01, 0x3f, 0xab, 0xb3, 0x07, 0x00, 0x00, 0xd4 },
      { 0x01, 0x21, 0x7f, 0xb3, 0x07, 0x00, 0x00, 0xdd },
      { 0x01, 0x5f, 0xb6, 0xb3, 0x07, 0x00, 0x00, 0x87 }
    },
    0.11, -2.6
  },
  {
    /* This one is marked "2" */
    { 0x26, 0x02, 0x89, 0x0d, 0x00, 0x00, 0x00, 0xaa },
    {
      { 0x01, 0xa7, 0xa3, 0xb3, 0x07, 0x00, 0x00, 0xfa },
      { 0x01, 0xbf, 0x9b, 0xb3, 0x07, 0x00, 0x00, 0xba },
      { 0x01, 0xa6, 0xe6, 0xb3, 0x07, 0x00, 0x00, 0xf6 },
      { 0x01, 0x70, 0xba, 0xb3, 0x07, 0x00, 0x00, 0x34 }
    },
    -0.45, 1.52
  },
  {
    /* This one is marked "3" */
    { 0x26, 0xdd, 0x4d, 0x09, 0x00, 0x00, 0x00, 0x79 },
    {
      { 0x01, 0x96, 0xd0, 0xb3, 0x07, 0x00, 0x00, 0x03 },
      { 0x01, 0xd4, 0x8d, 0xb3, 0x07, 0x00, 0x00, 0x61 },
      { 0x01, 0xee, 0xac, 0xb3, 0x07, 0x00, 0x00, 0x76 },
      { 0x01, 0x4b, 0xe2, 0xb3, 0x07, 0x00, 0x00, 0x8a }
    },
    -0.23, -0.42
  },
  {
    /* Spare thermo for outdoors.  */
    { 0x26, 0xff, 0x80, 0x0d, 0x00, 0x00, 0x00, 0x61 },
    {
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
    },
    0.51, 1.44
  }
};

FurnaceData furnace;
int fbits;

#define IDLEBIT 0
#define SERBIT(z) (2<<z)
#define OWBIT 0x01
#define OWSENSE 0x40

Var last_ac_on("last-ac-on");

#define LATCH_OW 1
#define LATCH_FZ 2
#define LATCH_NO 3

void latch(int which, int bits)
{
  outb(bits, 0x378);
  outb(which, 0x37a);
  upause(3);
  outb(LATCH_NO, 0x37a);
  if (which != LATCH_FZ)
    {
      outb(fbits, 0x378);
      outb(LATCH_FZ, 0x37a);
      upause(3);
      outb(LATCH_NO, 0x37a);
    }
}

int
reset()
{
  int i, r = 0;
  int did_idle = 0;
  int did_sense = 0;
  asm("cli");
  latch(LATCH_OW, OWBIT);
  zero_time();
  while (1)
  {
    i = elapsed_time();
    if (i >= 480 && !did_idle)
      {
        latch(LATCH_OW, IDLEBIT);
        did_idle = 1;
      }
    if (i >= 550 && !did_sense)
      {
        r=inb(0x379);
        did_sense = 1;
      }
    if (i >= 960)
      {
        asm("sti");
        return (r & OWSENSE) ? 0 : 1;
      }
  }
}

int
wbit(int b)
{
  int i, r;
  int did_idle = 0;
  int did_sense = 0;
  b = b ? 5 : 45;
  asm("cli");
  latch(LATCH_OW, OWBIT);
  zero_time();
  while (1)
  {
    i = elapsed_time();
    if (i >= b && !did_idle)
      {
        latch(LATCH_OW, IDLEBIT);
        did_idle = 1;
      }
    if (i >= 15 && !did_sense)
      {
        r = inb(0x379);
        did_sense = 1;
      }
    if (i >= 65 && did_idle)
      break;
  }
  asm("sti");
  return (r & OWSENSE) ? 1 : 0;
}

int
wr(int v)
{
  int i, rv=0, r;
  for (i=0; i<8; i++)
    {
      r = wbit(v & (1<<i));
      if (r)
	rv |= 1<<i;
    }
  return rv;
}

#define rd() wr(0xff)
#define rbit() (wbit(0xff) & 1)

void
port_init(void)
{
  iopl(3);
  fbits = 0xff;
  latch(LATCH_FZ, 0xff);
}

unsigned char onewirebuf[100];
unsigned char *onewireval;
static int rom_is_selected = 0;
unsigned char selected_rom[8];

void
select_rom(unsigned char *rom)
{
  if (rom == 0)
    rom_is_selected = 0;
  else
    {
      rom_is_selected = 1;
      memcpy (selected_rom, rom, 8);
    }
}

/*#define RETURN0 {printf("(oret %d)\n", __LINE__);return 0;}*/
#define RETURN0 return 0

void
wait_for_complete()
{
  /* Three in case we get a presence pulse.  */
  while (!rbit());
  while (!rbit());
  while (!rbit());
}

int
onewire(int len, ...)
{
  int i, rv = 1;
  int do_wait = 0;
  va_list a;

  onewireval = onewirebuf;

  if (len < 0)
    {
      len = -len;
      do_wait = 1;
    }

  if (reset() == 0)
    RETURN0;

  if (rom_is_selected)
    {
      int i = wr(0x55);
      if (i != 0x55)
	{
	  //printf("0x55 != 0x%x\n", i);
	  RETURN0;
	}
      for (i=0; i<8; i++)
	if (wr(selected_rom[i]) != selected_rom[i])
	  RETURN0;
    }
  else
    {
      if (wr(0xcc) != 0xcc)
	RETURN0;
    }

  va_start (a, len);
  for (i=0; i<len; i++)
    {
      int c = va_arg (a, int);
      if (c == -1)
	{
	  onewireval = onewirebuf + i;
	  break;
	}
      onewirebuf[i] = wr(c);
      if (onewirebuf[i] != c)
	rv = 0;
    }
  for (; i<len; i++)
    onewirebuf[i] = wr(0xff);
  if (do_wait)
    wait_for_complete();
  return rv;
}

static unsigned char crc_table[] = {
        0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
      157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
       35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
      190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
       70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
      219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
      101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
      248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
      140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
       17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
      175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
       50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
      202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
       87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
      233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
      116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53};

int
crc(unsigned char *data, int len)
{
  int crc = 0;
  while (len--)
    crc = crc_table[crc ^ *data++];
  return crc;
}

#define S_ZERO	0
#define S_ONE	1
#define S_TRY1	2

#define DEBUG	0

// Returns TRUE if the search was dirty (i.e. true means "try again")
int
search_rom(void (*func)(unsigned char *))
{
  unsigned char serno[8];
  int i, branch = -1;
  unsigned char bits[64];

#if DEBUG
  printf("%s: search rom!\n", ts());
#endif

  memset(bits, 0, sizeof(bits));

  while (1)
    {
    try_again:
      if (reset() == 0)
	{
	  printf("nothing on the bus\n");
	  return 0;
	}

      /* search rom command */
      wr(0xf0);

#if DEBUG
      printf("[b %2d] ", branch);
#endif

      // BRANCH points to the last bits[] that had S_TRY1.  Before
      // that, we pass through.  After that, we record new branches.
      // After the loop, we change the last S_TRY1 to S_ONE and point
      // BRANCH there.  Thus, The entry == BRANCH should be a pass
      // through.

      memset (serno, 0, sizeof(serno));
      for (i=0; i<64; i++)
	{
	  int bit;
	  int ibit;

	  bit = rbit();
	  ibit = rbit();
	  switch (bit + ibit*2)
	    {
	    case 0:
	      if (i <= branch)
		{
		  bit = bits[i] & 1;
#if DEBUG
		  putchar('A'+bit);
#endif
		}
	      else
		{
		  bit = 0;
		  bits[i] = S_TRY1;
#if DEBUG
		  putchar('a'+bit);
#endif
		}
	      break;

	    case 1:
	    case 2:
	      bits[i] = bit;
#if DEBUG
	      putchar('0'+bit);
#endif
	      if (i == 0 && bit == 0 && branch == -1)
		return 0;
	      break;

	    case 3:
#if DEBUG
	      printf("X\033[K\n");
#endif
	      return 1;
	    }

	  wbit(bit);
	  if (bit)
	    serno[i/8] |= 1<<(i&7);
	}
#if DEBUG
      printf("\033[K\n");
#endif
      if (crc(serno, 8) == 0)
	{
	  select_rom (serno);
	  func (serno);
	}
      else
	/* Bad crc, start again. */
	return 1;

      for (i=63; i>=0; i--)
	if (bits[i] == S_TRY1)
	  {
	    branch = i;
	    bits[i] = S_ONE;
	    break;
	  }
      if (i == -1)
	return 0;
    }
}

void
show_rom(unsigned char *rom)
{
  int z, b;
  for (z=0; z<3; z++)
    for (b=0; b<4; b++)
      if (memcmp (rom, Zone[z].button_ser[b], 8) == 0)
	{
	  Zone[z].buttons |= (1<<b);
	  return;
	}
}

#define SLOW_BUTTONPOLL 300
#define FAST_BUTTONPOLL 30
#define CLIMATE_TIMEOUT 3

time_t now;
time_t button_timeout;
time_t climate_timeout;

int
poll_interval(void)
{
  if (now - button_timeout < 0)
    return FAST_BUTTONPOLL;
  return SLOW_BUTTONPOLL;
}

void
poll_buttons(void)
{
  int z;
  int any_pressed = 0;

  for (z=0; z<3; z++)
    Zone[z].buttons = 0;

  if (search_rom(show_rom))
    return;

  for (z=0; z<3; z++)
    {
      any_pressed |= Zone[z].buttons;

      if (Zone[z].old_buttons != Zone[z].buttons)
	{
	  int b;
	  int changed = Zone[z].old_buttons ^ Zone[z].buttons;
	  int pressed = Zone[z].buttons;
	  int mask;

	  button_timeout = now + BUTTON_TIMEOUT;

	  for (b=0; b<4; b++)
	    {
	      mask = 1<<b;
	      if (changed & mask)
		{
		  ButtonData bd;
		  bd.zone = z;
		  bd.button = b;
		  bd.pressed = (pressed & mask) ? 1 : 0;
		  printf("%s zone %d button %d %s\n", ts(), z, b, bd.pressed ? "pressed" : "released");

		  Message msg (&bd, sizeof(bd), MT_button);
		  msg.send();
		}
	    }
	}
    }

  for (z=0; z<3; z++)
    Zone[z].old_buttons = Zone[z].buttons;
}

int
s16(unsigned char lsb, unsigned char msb)
{
  int rv;
  rv = lsb + (int)msb * 256;
  if (rv & 0x8000)
    rv -= 0x10000;
  return rv;
}

int last_zone = -1;

/* Returns nonzero if climate was successfully updated.  Stored in
   temps[zone][zindex[zone]] then zindex is incremented.  */
int
poll_climate (int first_time)
{
  int i = 0, p;
  int page[8][8];

  last_zone = (last_zone + 1) % 4;
  ZoneStruct *z = &(Zone[last_zone]);

/*#define RETURN {printf("(ret %d)\n", __LINE__);return 0;}*/
#define RETURN return 0

  select_rom (z->temp_ser);

  for (p=0; p<2; p++)
    {
      /* write page 0, Vad or Vdd */
      if (!onewire(3, 0x4e, 0x00, p&1 ? 0x08 : 0x00))
	RETURN;

      /* save page 0 */
      if (!onewire(-2, 0x48, 0x00))
	RETURN;

      if (p == 0)
	/* initiate T convert */
	if (!onewire(-1, 0x44))
	  RETURN;

      /* initiate V convert */
      if (!onewire(-1, 0xb4))
	RETURN;

      /* recall page */
      if (!onewire(2, 0xb8, 0))
	RETURN;

      /* read page */
      if (!onewire (11, 0xbe, 0, -1))
	RETURN;
      i = crc(onewireval, 9);
      if (i != 0)
	RETURN;
      for (i=0; i<8; i++)
	page[p][i] = onewireval[i];
    }

  int ti = s16(page[0][1], page[0][2]);
  double temp = (ti/256.0) * 1.8 + 32.0;

  int vad = s16(page[0][3], page[0][4]);
  int vdd = s16(page[1][3], page[1][4]);
  double humidity = ((((double)vad/(double)vdd) - 0.16) / 0.0062) / (1.093 - 0.0012 * temp);

  humidity += z->humidity_addend;
  temp += z->temp_addend;

  humidity = humidity / 1.15;
  if (humidity > 99.9)
    humidity = 99.9;
  if (humidity < 0)
    humidity = 0;

  int ai = z->avgidx;
  z->temp += (temp - z->temps[ai]) / NAVG;
  z->humidity += (humidity - z->humids[ai]) / NAVG;
  z->temps[ai] = temp;
  z->humids[ai] = humidity;
  z->avgidx = (z->avgidx + 1) % NAVG;

  if (first_time)
    {
      for (int i=1; i<NAVG; i++)
	{
	  z->temps[i] = z->temps[0];
	  z->humids[i] = z->humids[0];
	}
      z->temp = z->temps[0];
      z->humidity = z->humids[0];
    }

  //printf("zone %d temp %.2f humid %.2f\n", last_zone, temp, humidity);

  ClimateData cd;
  cd.zone = last_zone;
  cd.temperature = z->temp;
  cd.humidity = z->humidity;
  Message cdm(&cd, sizeof(cd), MT_climate);
  cdm.send();

  return 1;
}

static void
serial_byte(int zone, int val)
{
  int pin = 2 << zone;
  int o[10], b;
  o[0] = pin;
  for (b=0; b<8; b++)
    o[b+1] = val & (1<<b) ? 0 : pin;
  o[9] = 0;
  asm("cli");
  latch(LATCH_OW, IDLEBIT);
  zero_time();
  for (b=0; b<10; b++)
    {
      latch(LATCH_OW, o[b]);
      while (elapsed_time() < (b+1)*(1000000/9600))
        ;
    }
  asm("sti");
}

void
serial_string(int zone, char *s, int len)
{
  while (len--)
    serial_byte(zone, *s++);
}

static void
exit_proc(int)
{
  last_ac_on.set(0);
  port_init();
  exit(0);
}

main(int argc, char **argv)
{
  int z, b;

  Message::set_my_name("parallel");
  Message::init();

  port_init ();
  time(&now);
  button_timeout = now - 1;
  climate_timeout = now - 1;

  signal(SIGTERM, exit_proc);
  signal(SIGINT, exit_proc);

  for (z=0; z<4; z++)
    {
      int i = 10;
      Zone[z].avgidx = 0;
      while (i--)
	{
	  last_zone = z-1;
	  if (poll_climate(1))
	    break;
	}
    }

  Message::set_alarm (poll_interval());

  while (1)
    {
      Message *m = Message::next();

      time(&now);

      if (m)
	{
	  switch (m->type)
	    {
	    case MT_lcd:
	      serial_string(m->data[0], (char *)m->data+1, m->length-1);
	      break;
	    case MT_lcd_pause:
	      {
		short x = *(short *)m->data;
		usleep(x*1000);
	      }
	      break;
	    case MT_set_furnace:
	      {
		furnace = *(FurnaceData *)m->data;
		fbits = 0;
		if (furnace.heat)     fbits |= 0x01;
		if (furnace.cool)     fbits |= 0x02;
		if (furnace.fan)      fbits |= 0x04;
		if (furnace.humidify) fbits |= 0x08;
		if (furnace.zone[0])  fbits |= 0x10;
		if (furnace.zone[1])  fbits |= 0x20;
		if (furnace.zone[2])  fbits |= 0x40;
		if (furnace.zone[3])  fbits |= 0x80;
                fbits ^= 0x0f;
		asm("cli");
		latch(LATCH_FZ, fbits);
		asm("sti");
	      }
	      break;

	    case MT_ask_furnace:
	      {
		Message reply (&furnace, sizeof(furnace), MT_set_furnace);
		reply.dest = m->source;
		reply.send();
	      }
              break;
	    }
	}
      else
	{
	  poll_buttons ();
	  if (now - climate_timeout > 0)
	    {
	      poll_climate(0);
	      climate_timeout = now + CLIMATE_TIMEOUT;
	    }
	  Message::set_alarm (poll_interval());
	}
    }
}

void upause(unsigned x)
{
  x = x * MHz + rdtsc();
  while (rdtsc() < x)
    ;
}
