// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

FILE *in;
FILE *out_h, *out_c;
char *base_name;
int width, height;

unsigned char **data;

int
get_int ()
{
  int rv = 0;
  int ch;
  while (1)
    {
      ch = fgetc(in);
      if (ch == '#')
	{
	  while (fgetc(in) != '\n')
	    ;
	  continue;
	}
      if (isdigit (ch))
	{
	  rv = ch - '0';
	  while (isdigit (ch = fgetc(in)))
	    rv = rv * 10 + ch - '0';
	  return rv;
	}
    }
}

void
load_file (char *fname)
{
  unsigned char ch;
  int maxval, x, y;

  in = fopen(fname, "rb");
  if (!in)
    {
      perror(fname);
      exit(1);
    }

  if (fgetc(in) != 'P')
    {
      printf("%s: not a pnm file\n", fname);
      exit(1);
    }
  ch = fgetc(in);
  switch (ch)
    {
    case '5':
      width = get_int();
      height = get_int();
      maxval = get_int();
      if (maxval != 255)
	{
	  printf("%s: only 8 bit images supported\n", fname);
	  exit(1);
	}
      data = (unsigned char **) malloc (height * sizeof (unsigned char *));
      for (y=0; y<height; y++)
	{
	  data[y] = (unsigned char *) malloc (width * sizeof (unsigned char));
	  fread(data[y], 1, width, in);
	}
      break;
    default:
      printf("%s: unsupported pnm type P%c\n", fname, ch);
      exit(1);
    }
  printf("Load %s: width %d height %d\n", fname, width, height);
}

FILE *
baseopen(char *fmt)
{
  FILE *rv;
  char tmp[100];
  sprintf(tmp, fmt, base_name);
  rv = fopen(tmp, "wb");
  if (!rv)
    {
      perror(tmp);
      exit(1);
    }
  return rv;
}

void
copy_page (int px, int py, int pn)
{
  int x, y, b, i;
  int c = 39 + strlen(base_name);
  char comma = '{';
  fprintf(out_c, "static const unsigned char %s_%d[] = ", base_name, pn);
  for (y=0; y<32; y+=8)
    for (x=0; x<120; x++)
      {
	b = 0;
	for (i=0; i<8; i++)
	  if (data[py+y+i][px+x] < 128)
	    b |= 1 << i;
	if (b < 10)
	  c += 2;
	else if (b < 100)
	  c += 3;
	else
	  c += 4;
	fputc(comma, out_c);
	comma = ',';
	if (c > 75)
	  {
	    fprintf(out_c, "\n  ");
	    c = 2;
	  }
	fprintf(out_c, "%d", b);
      }
  fprintf(out_c, " };\n\n");
}

int
main(int argc, char **argv)
{
  int px, py, pn;
  if (argc < 3)
    {
      fprintf(stderr, "usage: pnm2page pnmfile base_name\n");
      exit(1);
    }
  load_file (argv[1]);
  base_name = argv[2];

  out_h = baseopen("%s.h");
  out_c = baseopen("%s.cc");

  fprintf(out_c, "#include \"%s.h\"\n\n", base_name);

  printf("Pages: %d by %d\n", width/120, height/32);

  pn = 0;
  for (py=0; py<height; py+=32)
    for (px=0; px<width; px+=120)
      copy_page (px, py, pn++);

  fprintf(out_c, "unsigned const char *%s[%d] = {\n", base_name, pn);
  for (px=0; px<pn; px++)
    fprintf(out_c, "  %s_%d%s\n", base_name, px, px<pn-1 ? "," : "");
  fprintf(out_c, "};\n");
  fprintf(out_c, "int %s_pages = %d;\n", base_name, pn);

  fprintf(out_h, "extern unsigned const char *%s[%d];\n", base_name, pn);
  fprintf(out_h, "extern int %s_pages;\n", base_name);
  fprintf(out_h, "#ifndef BYTES_PER_LCDPAGE\n");
  fprintf(out_h, "#define BYTES_PER_LCDPAGE (120*32/8)\n");
  fprintf(out_h, "#endif\n");

  fclose(out_h);
  fclose(out_c);
  exit(0);
}
