// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "msg.h"
#include "vars.h"

static Var *varlist = 0;

void
Var::set_name(const char *Pname)
{
  Var *v;
  if (name)
    {
      printf("attempt to change var %s name to %s\n", name, Pname);
      abort();
    }
  next = varlist;
  name = Pname;
  value = 0;
  multi = 0;

  for (v=varlist; v; v=v->next)
    if (strcmp (v->name, name) == 0)
      {
	v->multi = 1;
	multi = 1;
      }
  varlist = this;
}

void
Var::set_name (int zone, const char *Pname)
{
  char *new_name = new char[4+strlen(Pname)];
  sprintf(new_name, "z%d-%s", zone, Pname);
  set_name(new_name);
}

Var::Var()
{
  name = 0;
};

Var::Var(const char *Pname)
{
  set_name(Pname);
};

Var::Var(int zone, const char *Pname)
{
  set_name(zone, Pname);
}

void
Var::set(VarValueType newvalue)
{
  value = newvalue;

  SetVarValueData svd;
  svd.id = id;
  svd.value = value;
  Message m(&svd, sizeof(svd), MT_set_var_value);
  m.send();

  if (multi)
    {
      Var *v;
      for (v=0; v<varlist; v++)
	if (v->id == id && v != this)
	  v->value = value;
    }
}

void
Var::setup_names()
{
  Var *v;
  for (v=varlist; v; v=v->next)
    {
      if (!v->name)
	{
	  printf("Error: Var at %x has no name, in setup_names\n", v);
	  abort();
	}
      if (v->id == VAR_NO_ID)
	{
	  Message m(v->name, strlen(v->name)+1, MT_ask_var_id);
	  m.send();
	}
    }
}

void
Var::wait_for_all_ids (void)
{
  Message *m;
  Var *v;
  int waiting = 1;

  setup_names();
  Message::set_alarm(1000);
  while (waiting)
    {
      m = Message::next();
      if (!m)
	{
	  setup_names();
	  Message::set_alarm(1000);
	  continue;
	}
      if (m->type == MT_var_id)
	{
	  Message::set_alarm(1000);
	  waiting = 0;
	  for (v=varlist; v; v=v->next)
	    if (v->id == VAR_NO_ID)
	      {
		waiting = 1;
		break;
	      }
	}
    }
  Message::set_alarm();
}

void
Var::handle_message(Message *m)
{
  Var *v;

  switch (m->type)
    {
    case MT_var_id:
      {
	VarIdData *vid = (VarIdData *)m->data;
	for (v=varlist; v; v=v->next)
	  if (strcmp (v->name, vid->name) == 0)
	    {
	      v->id = vid->id;
	      v->value = vid->value;
		if (!v->multi)
		  return;
	    }
	break;
      }

    case MT_set_var_value:
      {
	SetVarValueData *svd = (SetVarValueData *)m->data;
	for (v=varlist; v; v=v->next)
	  if (v->id == svd->id)
	    {
	      v->value = svd->value;
	      if (!v->multi)
		return;
	    }
	break;
      }
    }
}
