// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/DrawingA.h>

#include "msg.h"
#include "furnace.h"

#define MAX 1600

#define MAXARGS 40
#define stdarg(p, v) XtSetArg(args[n], p, v), n++
int n;
Arg args[MAXARGS];

int test = 1;

FurnaceData fd[MAX];
double temp[MAX][4];
time_t times[MAX];
int nd = 0;

struct {
  double temp;
  double humidity;
} zone[4];

char *cmd;

Widget da;
GC off[4], on[4];
GC heatbg_gc, coolbg_gc, fanbg_gc, humidifybg_gc;
GC bg_light_gc, bg_dark_gc, bg_gc;
GC time_light_gc, time_dark_gc, time_gc;
GC notset_gc;

Display *display = 0;
int screen;
Window daw;
Pixmap dap = 0;
int pix_w, pix_h;
Colormap colormap;

static int time_marker, old_time_marker = -1;

time_t
xtime()
{
  struct timeval tv;
  struct tm *tm;
  gettimeofday(&tv, 0);
  tm = localtime(&tv.tv_sec);
  if (test)
    time_marker = tv.tv_sec / 5;
  else
    time_marker = tv.tv_sec / 3600;

  if (test ? (tm->tm_sec < 15 || tm->tm_sec > 45)
      : (tm->tm_hour < 6 || tm->tm_hour > 18))
    {
      time_gc = time_dark_gc;
      bg_gc = bg_dark_gc;
    }
  else
    {
      time_gc = time_light_gc;
      bg_gc = bg_light_gc;
    }
  if (old_time_marker == -1)
    old_time_marker = time_marker;
  if (test)
    return tv.tv_sec*10 + (tv.tv_usec / 100000);
  return tv.tv_sec / 60;
}

GC
make_gc(int r, int g, int b)
{
  GC gc = XCreateGC(display, daw, 0, 0);
  XColor c;
  c.red   = r * 0x0101;
  c.green = g * 0x0101;
  c.blue  = b * 0x0101;
  c.flags = DoRed | DoGreen | DoBlue;
  if (XAllocColor (display, colormap, &c))
    XSetForeground(display, gc, c.pixel);
  else
    XSetForeground(display, gc, BlackPixel(display, screen));
  XSetBackground(display, gc, WhitePixel(display, screen));
  return gc;
}

void
plot (void)
{
  int i, z;
  GC gc, bg;

  if (!daw)
    return;

  int pi = (nd+MAX-1) % MAX;
  int ti = nd;

  if (fd[pi].heat)
    bg = fd[pi].humidify ? humidifybg_gc : heatbg_gc;
  else if (fd[pi].cool)
    bg = coolbg_gc;
  else if (fd[pi].fan)
    bg = fanbg_gc;
  else 
    bg = bg_gc;

  time_t now = times[ti];
  int dx = now - times[pi];
  if (dx)
    {
      int td = pix_h/2 / 100;
      XCopyArea(display, dap, dap, bg_gc, dx, 0, pix_w-dx, pix_h, 0, 0);
      XFillRectangle(display, dap, times[pi] ? bg : notset_gc, pix_w-dx, 0, dx, pix_h);
      if (time_marker != old_time_marker)
	{
	  old_time_marker = time_marker;
	  XDrawLine (display, dap, time_gc, pix_w-1, 0, pix_w-1, pix_h);
	}
      for (z=-td; z<=td; z++)
	XDrawPoint(display, dap, time_gc, pix_w-1, pix_h/2 + z*100);
    }

  for (z=0; z<4; z++)
    {
      gc = (bg != bg_gc && fd[pi].zone[z]) ? on[z] : off[z];

      if (temp[pi][z])
	XDrawLine (display, dap, gc,
		   pix_w-1-dx, pix_h/2 + (int)(700-temp[pi][z]*10),
		   pix_w-1,    pix_h/2 + (int)(700-temp[ti][z]*10));
    }


  XCopyArea(display, dap, daw, bg_gc, 0, 0, pix_w, pix_h, 0, 0);
}

void
expose_callback(Widget w, void *, XmDrawingAreaCallbackStruct *cbs)
{
  Dimension width;
  Dimension height;
  XExposeEvent *e = &(cbs->event->xexpose);

  if (!display)
    {
      daw = XtWindow(w);
      display = XtDisplay(da);
      screen = XDefaultScreen(display);
      colormap = XDefaultColormap(display, screen);

      notset_gc = make_gc(50, 30, 10);
      bg_gc = bg_dark_gc = make_gc(0, 0, 0);
      bg_light_gc = make_gc(0, 10, 20);
      time_gc = time_dark_gc = make_gc(30, 30, 30);
      time_light_gc = make_gc(60, 60, 60);

      heatbg_gc = make_gc(50, 0, 0);
      humidifybg_gc = make_gc(50, 50, 0);
      coolbg_gc = make_gc(0, 0, 50);
      fanbg_gc = make_gc(0, 50, 0);

      on[0] = make_gc(255, 50, 50);
      on[1] = make_gc(50, 255, 50);
      on[2] = make_gc(50, 128, 255);
      off[3] = make_gc(200, 200, 0);
      off[0] = make_gc(100, 0, 0);
      off[1] = make_gc(0, 100, 0);
      off[2] = make_gc(0, 50, 100);
      off[3] = make_gc(100, 100, 0);
    }

  n = 0;
  stdarg(XmNwidth, &width);
  stdarg(XmNheight, &height);
  XtGetValues(da, args, n);

  if (!dap || pix_w != width || pix_h != height)
    {
      if (dap)
	XFreePixmap(display, dap);
      dap = XCreatePixmap(display, daw, width, height, XDefaultDepth(display, screen));
      pix_w = width;
      pix_h = height;
      XFillRectangle(display, dap, notset_gc, 0, 0, width, height);
      XFillRectangle(display, daw, notset_gc, 0, 0, width, height);
    }

  XCopyArea(display, dap, daw, bg_gc, e->x, e->y, e->width, e->height, e->x, e->y);
}

FurnaceData saved_furnace;

static void
plot_now(time_t now)
{
  if (times[nd] == now)
    return;
  nd = (nd+1) % MAX;
  fd[nd] = saved_furnace;
  temp[nd][0] = zone[0].temp;
  temp[nd][1] = zone[1].temp;
  temp[nd][2] = zone[2].temp;
  temp[nd][3] = zone[3].temp;
  times[nd] = now;
  plot();
}

static void
xmp_read_message (void *ptr, int *source, XtInputId *id)
{
  Message *m;

  while (1)
    {
      m = Message::next(0);
      if (!m)
	return;

      time_t now = xtime();

      switch (m->type)
	{
	case MT_set_furnace:
	  {
	    if (zone[0].temp && zone[1].temp && zone[2].temp)
	      {
		saved_furnace = *(FurnaceData *)m->data;
		plot_now(now);
	      }
	    break;
	  }

	case MT_climate:
	  {
	    ClimateData *cd = (ClimateData *)m->data;
	    if (cd->zone <= 3)
	      {
		zone[cd->zone].temp = cd->temperature;
		zone[cd->zone].humidity = cd->humidity;
	      }
	    plot_now(now);
	    break;
	  }
	}
    }
}

int
main(int argc, char **argv)
{
  Message *m;

  if (argc > 1)
    {
      test = 0;
      Message::set_host(argv[1]);
    }

  Message::set_my_name("zplotter");
  Message::init();

  Widget toplevel = XtInitialize(0, "XmPlotter",
				 0, 0,
				 &argc, argv);

  n = 0;
  stdarg(XmNresizePolicy, XmRESIZE_ANY);
  stdarg(XmNshadowThickness, 0);
  Widget form_widget = XmCreateForm(toplevel, "form", args, n);
  XtManageChild(form_widget);

  n = 0;
  stdarg(XmNtopAttachment, XmATTACH_FORM);
  stdarg(XmNbottomAttachment, XmATTACH_FORM);
  stdarg(XmNleftAttachment, XmATTACH_FORM);
  stdarg(XmNrightAttachment, XmATTACH_FORM);
  stdarg(XmNwidth, 500);
  stdarg(XmNheight, (90-50)*10);
  stdarg(XmNresizePolicy, XmRESIZE_ANY);
  da = XmCreateDrawingArea(form_widget, "plot", args, n);
  XtManageChild(da);
  XtAddCallback(da, XmNexposeCallback, (XtCallbackProc)expose_callback, 0);

  XtAddInput (Message::mux_socket(), (XtPointer)XtInputReadMask,
	      (XtInputCallbackProc)xmp_read_message, 0);
  XtRealizeWidget(toplevel);
  XtMainLoop();
}
