// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "msg.h"
#include "vars.h"
#include "misc.h"
#include "furnace.h"

#define MAX 1000

FurnaceData fd[MAX];
double temp[MAX][3];
time_t times[MAX];
int n = 0;

double tmax=90, tmin=50;

struct {
  double temp;
  double humidity;
} zone[3];

char *cmd;

void
color(FILE *ps, int r, int g, int b, int x, int y)
{
  static int sr=-1, sg, sb;
  if (sr == r && sg == g && sb == b)
    return;
  sr = r;
  sg = g;
  sb = b;
  if (ps == 0)
    return;
  if (strcmp(cmd, "moveto"))
    fprintf(ps, "%d %d %s stroke\n", x, y, cmd);
  fprintf(ps, "%d %d %d setrgbcolor\n", r, g, b);
  cmd = "moveto";
}

void
plot (void)
{
  int i, z;
  FILE *ps = fopen("temps.ps", "w");
  fprintf(ps, "%!PS-Adobe\n");
  fprintf(ps, "36 36 translate\n");
  
  for (z=0; z<3; z++)
    {
      cmd = "moveto";
      color(0, -1, 0, 0, 0, 0);
      for (i=0; i<n; i++)
	{
	  int x = (int)(i * 7.5 * 72 / n);
	  int y = (int)((temp[i][z] - tmin) / (tmax - tmin) * 10 * 72);
	  int cr=0, cg=0, cb=0;
	  if (fd[i].heat && fd[i].zone[z])
	    cr = 1;
	  else if (fd[i].cool && fd[i].zone[z])
	    cb = 1;
	  else if (fd[i].fan && fd[i].zone[z])
	    cg = 1;
	  color(ps, cr, cg, cb, x, y);
	  fprintf(ps, "%d %d %s\n", x, y, cmd);
	  cmd = "lineto";
	}
      fprintf(ps, "stroke\n");
    }
  fprintf(ps, "showpage\n");
  fclose(ps);
}

int
main(int argc, char **argv)
{
  Message *m;
  time_t now, then;

  Message::set_my_name("zplotter");
  Message::init();

  while (1)
    {
      m = Message::next();
      time(&now);

      switch (m->type)
	{
	case MT_set_furnace:
	  {
	    if (zone[0].temp && zone[1].temp && zone[2].temp)
	      {
		fd[n] = *(FurnaceData *)m->data;
		temp[n][0] = zone[0].temp;
		temp[n][1] = zone[1].temp;
		temp[n][2] = zone[2].temp;
		times[n] = now;
		n++;
		if (n >= MAX)
		  exit(0);
		if (now - then > 5)
		  {
		    plot();
		    then = now;
		  }
	      }
	    break;
	  }

	case MT_climate:
	  {
	    ClimateData *cd = (ClimateData *)m->data;
	    if (cd->zone <= 2)
	      {
		zone[cd->zone].temp = cd->temperature;
		zone[cd->zone].humidity = cd->humidity;

		if (tmax < cd->temperature)
		  tmax = cd->temperature;
		if (tmin > cd->temperature)
		  tmin = cd->temperature;
	      }
	    break;
	  }
	}
    }
}
